#include <limits.h>
#include <stdlib.h>
#include <errno.h>
#include <libspe2.h>
#include <time.h>
#include <stdio.h>
#include <pthread.h>
#include <math.h>

// Includes ppu/spu data structs, and stub.h (from spiral).
#include "speinfo.h"

__attribute__ ((aligned(128))) DATATYPE_NO_QUOTES X[SIZE];
__attribute__ ((aligned(128))) DATATYPE_NO_QUOTES Y[SIZE];

extern spe_program_handle_t spiral_spu;


void *spe_perform_timing(void *);

void * inc_on_spe(void *vptr_args) {
  spu_t* spu_thread = (spu_t*) vptr_args;
  spe_stop_info_t stopinfo;
  unsigned int entry = SPE_DEFAULT_ENTRY;
  spe_context_run(spu_thread->context, &entry, 0, &(spu_thread->spe_info), 0, &stopinfo);
  if (stopinfo.stop_reason == SPE_EXIT) {
    if (stopinfo.result.spe_exit_code != 0) {
      fprintf(stderr, "ERROR: SPU failed with error code %d\n", stopinfo.result.spe_exit_code);
      exit(1);
    }
  }
  else {
    fprintf(stderr, "ERROR: SPU failed with unknown error. Go figure.\n");
    exit(1);
  }
  pthread_exit(NULL);
}

double perform_timing(int num_runs) {

  // fill input data into aligned array
  for(int j = 0; j < SIZE; j++)
    X[j] = 0;

  spu_t spu_thread[SPUS];

  // Create SPU threads
  for (int i=0; i<SPUS; i++) {
    // Create SPU contexts
    spu_thread[i].context = spe_context_create((SPE_MAP_PS|SPE_CFG_SIGNOTIFY1_OR), 0);
    spu_thread[i].spe_info.X = (unsigned long long) &X[0];
    spu_thread[i].spe_info.Y = (unsigned long long) &Y[0];
    spu_thread[i].spe_info.spuid = i;
    spu_thread[i].spe_info.num_runs = num_runs;
    spu_thread[i].spe_info.timing_mode = 1;

    printf("PPE: addr of X = %llu\n", spu_thread[i].spe_info.X);
    printf("PPE: addr of Y = %llu\n", spu_thread[i].spe_info.Y);

    // Load all other SPUs' info to this SPU's infostruct
    for (int j=0; j<SPUS; j++) {
      spu_thread[i].spe_info.spu_ls[j] = 
        (unsigned long long) spe_ls_area_get(spu_thread[j].context);
      spu_thread[i].spe_info.spu_ps[j] = 
        (unsigned long long) spe_ps_area_get(spu_thread[j].context, SPE_SIG_NOTIFY_1_AREA);
    }

    spe_program_load(spu_thread[i].context, &spiral_spu);
    pthread_create(&(spu_thread[i].sputhreadt), NULL, inc_on_spe, &spu_thread[i]);
    debug_ppe("PPU: Creating thread %2d\n", i);
  }

  for (int i=0; i<SPUS; i++) {
    debug_ppe("PPU: LS_Area[%d]   = %llu\n", i, spu_thread[0].spe_info.spu_ls[i]);
    debug_ppe("PPU: Sig1_Area[%d] = %llu\n", i, spu_thread[0].spe_info.spu_ps[i]);
  }

  // Send the spuids to the SPUs, and start them.
  for (unsigned int i=0; i<SPUS; i++) {
    spe_in_mbox_write(spu_thread[i].context, &i, 1, SPE_MBOX_ALL_BLOCKING);
  }

  // Join
  for (int i=0; i<SPUS; i++) {
    pthread_join(spu_thread[i].sputhreadt, NULL);
    spe_context_destroy(spu_thread[i].context);
  }

  /* get the maximum cycle count (max over SPUs) */

  unsigned int cycles, max_cycles = 0;
  for(int i=0; i<SPUS; i++) {
    cycles = spu_thread[i].spe_info.cycles;
    debug_ppe("PPU: SPU %u's cycle count was %u\n", i, cycles);
    if(cycles > max_cycles)
      max_cycles = cycles;
  }
  debug_ppe("PPU: max_cycles was %.0f\n", (float) max_cycles*CYCLES_PER_TIMEBASE);

  cycles = 0;
  unsigned int min_cycles = max_cycles;
  for(int i=0; i<SPUS; i++) {
    cycles = spu_thread[i].spe_info.cycles;
    debug_ppe("PPU: SPU %u's cycle count was %u\n", i, cycles);
    if(cycles < min_cycles)
      min_cycles = cycles;
  }

  debug_ppe("PPU: min_cycles was %.0f\n", min_cycles*CYCLES_PER_TIMEBASE);

#ifdef TIME_MAX_CYCLES
  return(max_cycles);
#else
  return(min_cycles);
#endif

}


int main(int argc, char** argv) {
  static double  result = 0;
  static double cycles, gflops;
  int num_runs=1;

  debug_ppe("PPU: start. SPUS=%d\n", SPUS);

  result = 0;
  while (result < TIMEBASE_MIN_PRECISION) {
    result = perform_timing(num_runs);
    num_runs *= 2;
  }
  num_runs /= 2;

  debug_ppe("result = %f", result);
  cycles = (float) (result / num_runs) * CYCLES_PER_TIMEBASE;
  gflops = (5*(SIZE/2)*log(SIZE/2)/log(2))/(cycles/FREQ_IN_GHZ);
  printf("Spiral DFT. Size %5d. %6.1f cycles. %6.2f Gflop/s\n", (SIZE/2), cycles, gflops);


  return 0;
}


